/***
 * Copyright (C) 1997,98 -- Miroslav Zagorac  <zaga@fly.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "lib_atexit.h"
#include "lib_debug.h"
#include "lib_image.h"
#include "lib_pcx.h"
#include "lib_string.h"
#include "lib_X11.h"


static int (*X11_original_error_handler) (Display *, XErrorEvent *) = NULL;

static Atom       wm_delete_window;
static palette_t *x11_palette = NULL;

Display          *x11_display = NULL;
int               x11_screen, x11_depth, x11_shm_completion;
XFontStruct      *x11_font_info = NULL;
Window            x11_win_root;
Colormap          x11_default_colormap, x11_private_colormap;
ulong             Pixel_color[X11_MAX_COLORS];
XVisualInfo       x11_visual_info;
XKeyboardState    x11_kbd_state;
short             x11_bpp, x11_shm_busy;

void (*X11_callback_keyevent) (XKeyEvent *event) = NULL;
void (*X11_callback_anyevent) (XEvent *event)    = NULL;


#ifdef DEBUG

#ifdef ANSI_FUNC

static char *X11_strerror (int error_code)
#else

static char *X11_strerror (error_code)
   int error_code;
#endif
{
   static char error_text[BUFSIZ];

   if (x11_display)
      XGetErrorText (x11_display, error_code, error_text, BUFSIZ);
   else
      (void) strcpy (error_text, "unknown");

   return (error_text);
}


#ifdef ANSI_FUNC

static char *X11_event_name (int type)
#else
   
static char *X11_event_name (type)
   int type;
#endif
{
   static char *event_name[LASTEvent] =
   {
      "error protocol",
      "replies protocol",
      "KeyPress",
      "KeyRelease",
      "ButtonPress",
      "ButtonRelease",
      "MotionNotify",
      "EnterNotify",
      "LeaveNotify",
      "FocusIn",
      "FocusOut",
      "KeymapNotify",
      "Expose",
      "GraphicsExpose",
      "NoExpose",
      "VisibilityNotify",
      "CreateNotify",
      "DestroyNotify",
      "UnmapNotify",
      "MapNotify",
      "MapRequest",
      "ReparentNotify",
      "ConfigureNotify",
      "ConfigureRequest",
      "GravityNotify",
      "ResizeRequest",
      "CirculateNotify",
      "CirculateRequest",
      "PropertyNotify",
      "SelectionClear",
      "SelectionRequest",
      "SelectionNotify",
      "ColormapNotify",
      "ClientMessage",
      "MappingNotify"
   };

   if ((type >= 0) && (type < LASTEvent))
      return (event_name[type]);

   return ("Incorrect event type!");
}

#endif /* DEBUG */


#ifdef ANSI_FUNC

static int X11_error_handler (Display *display, XErrorEvent *error_event)
#else

static int X11_error_handler (display, error_event)
   Display *display;
   XErrorEvent *error_event;
#endif
{
   debug (("{%s(0x%08lx, 0x%08lx)", __FUNCTION__, (long) display, (long) error_event))

   if (X11_original_error_handler)
      X11_original_error_handler (display, error_event);

   debug_return ("} == %d", 0)
}


#ifdef ANSI_FUNC

short X11_init (void)
#else

short X11_init ()
#endif
{
   debug (("{%s()", __FUNCTION__))

   if (! (x11_display = XOpenDisplay (getenv ("DISPLAY"))))
   {
      debug (("XOpenDisplay(): %s\n", X11_strerror (0)))

      debug_return ("} == %d", FUNC_RET_ERROR)
   }

   /***
    * Get current control values for the keyboard.
    */
   XGetKeyboardControl (x11_display, &x11_kbd_state);

   XSetCloseDownMode (x11_display, DestroyAll);

   X11_original_error_handler = XSetErrorHandler (X11_error_handler);

   x11_screen           = XDefaultScreen (x11_display);
   x11_win_root         = XRootWindow (x11_display, x11_screen);
   x11_depth            = XDefaultDepth (x11_display, x11_screen);
   x11_default_colormap = XDefaultColormap (x11_display, x11_screen);

   if (XMatchVisualInfo (x11_display, x11_screen, 8, PseudoColor, &x11_visual_info))
      x11_bpp = 1;
   else if (XMatchVisualInfo (x11_display, x11_screen, 15, TrueColor, &x11_visual_info))
      x11_bpp = 2;
   else if (XMatchVisualInfo (x11_display, x11_screen, 16, TrueColor, &x11_visual_info))
      x11_bpp = 2;
   else if (XMatchVisualInfo (x11_display, x11_screen, 24, TrueColor, &x11_visual_info))
      x11_bpp = 3;
   else if (XMatchVisualInfo (x11_display, x11_screen, 32, TrueColor, &x11_visual_info))
      x11_bpp = 4;
   else
   {
      debug (("Screen doesn't support PseudoColor or TrueColor!\n"))

      debug_return ("} == %d", FUNC_RET_ERROR)
   }

   debug (("x11_visual_info = {\n"))
   debug (("   visual        = 0x%08lx\n", (long) x11_visual_info.visual))
   debug (("   visualid      = 0x%08x\n", (int) x11_visual_info.visualid))
   debug (("   screen        = %d\n", x11_visual_info.screen))
   debug (("   depth         = %d\n", x11_visual_info.depth))
   debug (("   class         = %d\n", x11_visual_info.class))
   debug (("   red_mask      = 0x%08lx\n", x11_visual_info.red_mask))
   debug (("   green_mask    = 0x%08lx\n", x11_visual_info.green_mask))
   debug (("   blue_mask     = 0x%08lx\n", x11_visual_info.blue_mask))
   debug (("   colormap_size = %ld\n", x11_visual_info.colormap_size))
   debug (("   bits_per_rgb  = %ld\n", x11_visual_info.bits_per_rgb))
   debug ((" }\n"))
   debug (("x11_bpp = %d\n", x11_bpp))
   debug (("DisplayPlanes()  = %d\n", DisplayPlanes (x11_display, x11_screen)))

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef ANSI_FUNC

void X11_close (void)
#else

void X11_close ()
#endif
{
   debug (("{%s()", __FUNCTION__))

   if (x11_display)
   {
      if (x11_private_colormap)
         XFreeColormap (x11_display, x11_private_colormap);

      if (x11_font_info)
         XFreeFont (x11_display, x11_font_info);

      if (x11_kbd_state.global_auto_repeat != AutoRepeatModeOff)
         XAutoRepeatOn (x11_display);

      XCloseDisplay (x11_display);
   }

   x11_display          = NULL;
   x11_font_info        = NULL;
   x11_private_colormap = 0;

   debug_endfunc
}


#ifdef ANSI_FUNC

void X11_handle_events (void)
#else
   
void X11_handle_events ()
#endif
{
   static XEvent event;
   long          qevents;

   if ((qevents = XPending (x11_display)))
   {
      debug (("{%s()", __FUNCTION__))

      XNextEvent (x11_display, &event);

      debug (("queued events: %ld\n", qevents))
      debug (("event_type:    %s  <%d>\n", X11_event_name (event.type), event.type))

      switch (event.type)
      {
         case ClientMessage : if ((event.xclient.format == 32) && event.xclient.data.l[0] == wm_delete_window)
                                 Exit (0);

                              break;

         case UnmapNotify   : while (event.type != MapNotify)
                              {
#ifdef HAVE_NANOSLEEP
                                 struct timespec req = { 0L, 50000000L };

                                 (void) nanosleep (&req, NULL);
#else
                                 (void) usleep (50000L);
#endif
                                 X11_handle_events ();
                              }

                              break;

         case EnterNotify   : XAutoRepeatOff (x11_display);

                              break;

         case LeaveNotify   : if (x11_kbd_state.global_auto_repeat != AutoRepeatModeOff)
                                 XAutoRepeatOn (x11_display);

                              break;

         case KeyPress      :
         case KeyRelease    : if (X11_callback_keyevent)
                                 X11_callback_keyevent ((XKeyEvent *) &event);

                              break;

         default            : if (event.type == x11_shm_completion)
                                 x11_shm_busy = 0;
                              else if (X11_callback_anyevent)
                                 X11_callback_anyevent (&event);
      }

      debug_endfunc
   }
}


#ifdef ANSI_FUNC

short X11_set_color (ulong red, ulong green, ulong blue, short color_index)
#else
   
short X11_set_color (ulong red, ulong green, ulong blue, short color_index)
   ulong red;
   ulong green;
   ulong blue;
   short color_index;
#endif
{
   short retval = FUNC_RET_OK;

   debug (("{%s(0x%08x, 0x%08x, 0x%08x, %d)", __FUNCTION__, red, green, blue, color_index))

   if ((color_index < 0) || (color_index >= X11_MAX_COLORS))
      debug_return ("} == %d", FUNC_RET_ERROR)

   if (x11_bpp == 1)
   {
      XColor xcolor;

      if (x11_palette)
      {
         x11_palette[color_index * 3]     = red;
         x11_palette[color_index * 3 + 1] = green;
         x11_palette[color_index * 3 + 2] = blue;
      }

      xcolor.pixel = color_index;
      xcolor.red   = red << 8;
      xcolor.green = green << 8;
      xcolor.blue  = blue << 8;
      xcolor.flags = DoRed | DoGreen | DoBlue;

      XStoreColor (x11_display, x11_private_colormap, &xcolor);
   }
   else
   {
      short red_shift   = fls (x11_visual_info.red_mask) - 8;
      short green_shift = fls (x11_visual_info.green_mask) - 8;
      short blue_shift  = fls (x11_visual_info.blue_mask) - 8;

      /***
       * Iskopirano if funkcije X11_set_colormap().
       */
      Pixel_color[color_index]  = ((red << ((red_shift > 0) ? red_shift : 0)) >> ((red_shift < 0) ? -red_shift : 0)) & x11_visual_info.red_mask;
      Pixel_color[color_index] |= ((green << ((green_shift > 0) ? green_shift : 0)) >> ((green_shift < 0) ? -green_shift : 0)) & x11_visual_info.green_mask;
      Pixel_color[color_index] |= ((blue << ((blue_shift > 0) ? blue_shift : 0)) >> ((blue_shift < 0) ? -blue_shift : 0)) & x11_visual_info.blue_mask;
   }

   debug_return ("} == %d", retval)
}


#ifdef ANSI_FUNC

short X11_set_colormap (palette_t *palette, short ncolors)
#else
   
short X11_set_colormap (palette, ncolors)
   palette_t *palette;
   short ncolors;
#endif
{
   short i;

   debug (("{%s(0x%08lx, %d)", __FUNCTION__, (long) palette, ncolors))

   x11_palette = palette;

   if (x11_bpp == 1)
   {
      /***
       * Allocate private colormap.
       */
      XColor xcolors[X11_MAX_COLORS];

      x11_private_colormap = XCreateColormap (x11_display, x11_win_root, XDefaultVisual (x11_display, x11_screen), AllocAll);

      for (i = 0; i < ncolors; i++)
      {
         xcolors[i].pixel = i;
         xcolors[i].red   = palette[i * 3] << 8;
         xcolors[i].green = palette[i * 3 + 1] << 8;
         xcolors[i].blue  = palette[i * 3 + 2] << 8;
         xcolors[i].flags = DoRed | DoGreen | DoBlue;
      }

      XStoreColors (x11_display, x11_private_colormap, xcolors, ncolors);
      XInstallColormap (x11_display, x11_private_colormap);
      XSetWindowColormap (x11_display, x11_win_root, x11_private_colormap);
   }
   else
   {
      short red_shift   = fls (x11_visual_info.red_mask) - 8;
      short green_shift = fls (x11_visual_info.green_mask) - 8;
      short blue_shift  = fls (x11_visual_info.blue_mask) - 8;

      /***
       * TrueColor. E, to je velika magija... B)
       */
      for (i = 0; i < ncolors; i++)
      {
         Pixel_color[i]  = ((palette[i * 3] << ((red_shift > 0) ? red_shift : 0)) >> ((red_shift < 0) ? -red_shift : 0)) & x11_visual_info.red_mask;
         Pixel_color[i] |= ((palette[i * 3 + 1] << ((green_shift > 0) ? green_shift : 0)) >> ((green_shift < 0) ? -green_shift : 0)) & x11_visual_info.green_mask;
         Pixel_color[i] |= ((palette[i * 3 + 2] << ((blue_shift > 0) ? blue_shift : 0)) >> ((blue_shift < 0) ? -blue_shift : 0)) & x11_visual_info.blue_mask;
      }
   }

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef ANSI_FUNC

void X11_colormap_fade (short direction)
#else
   
void X11_colormap_fade (direction)
   short direction;
#endif
{
   debug (("{%s(%d)", __FUNCTION__, direction))

   debug_endfunc
}


#ifdef ANSI_FUNC

short X11_window_open (struct WINDOW_DATA *win_data, short width, short height)
#else

short X11_window_open (win_data, width, height)
   struct WINDOW_DATA *win_data;
   short width;
   short height;
#endif
{
   XSizeHints sizehints;
   XWMHints   wmhints;
   XClassHint classhints;

   debug (("{%s(0x%08lx, %d, %d)", __FUNCTION__, (long) win_data, width, height))

   if (! win_data)
      debug_return ("} == %d", FUNC_RET_ERROR);

   /***
    * Nedam moj prozor da mi netko nateze...
    */
   sizehints.flags  = USSize | PMinSize | PMaxSize;
   sizehints.width  = sizehints.min_width  = sizehints.max_width  = width;
   sizehints.height = sizehints.min_height = sizehints.max_height = height;

   wmhints.initial_state = NormalState;
   wmhints.input         = True;
   wmhints.flags         = StateHint | InputHint;

   classhints.res_name  = win_data->window_name;
   classhints.res_class = win_data->window_name;

   if (! (win_data->window = XCreateWindow (x11_display, x11_win_root, 0, 0, sizehints.width, sizehints.height, 2, x11_depth, InputOutput, CopyFromParent, win_data->attr_mask, &win_data->attr)))
   {
      debug (("XCreateWindow(): %s\n", X11_strerror (0)))

      debug_return ("} == %d", FUNC_RET_ERROR)
   }

   if (win_data->icon)
   {
#ifdef HAVE_LIBXPM
      XpmAttributes xpm_attr;
      Pixmap        icon_pixmap;

      xpm_attr.valuemask = XpmColormap;
      xpm_attr.colormap  = x11_default_colormap;

      if (XpmCreatePixmapFromData (x11_display, win_data->window, win_data->icon, &icon_pixmap, NULL, &xpm_attr) != XpmSuccess)
         debug (("XpmCreatePixmapFromData(): %s\n", X11_strerror (0)))

      XpmFreeAttributes (&xpm_attr);
      
      wmhints.icon_pixmap  = icon_pixmap;
      wmhints.flags       |= IconPixmapHint;
#else
#endif /* HAVE_LIBXPM */
   }

   /***
    * Create the Graphics Contexts structure.
    */
   win_data->gc = XCreateGC (x11_display, win_data->window, win_data->gc_values_mask, &win_data->gc_values);

   /***
    * Enable the delete window protocol (taken from 'rxvt' -- Thanks!)
    */
   wm_delete_window = XInternAtom (x11_display, "WM_DELETE_WINDOW", False);
   (void) XSetWMProtocols (x11_display, win_data->window, &wm_delete_window, 1);

   XmbSetWMProperties (x11_display, win_data->window, win_data->window_name, win_data->icon_name, NULL, 0, &sizehints, &wmhints, &classhints);
   (void) XSelectInput (x11_display, win_data->window, win_data->event_mask);
   (void) XMapWindow (x11_display, win_data->window);

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef ANSI_FUNC

void X11_window_close (struct WINDOW_DATA *win_data)
#else
   
void X11_window_close (win_data)
   struct WINDOW_DATA *win_data;
#endif
{
   debug (("{%s(0x%08lx)", __FUNCTION__, (long) win_data))

   /***
    * Unmap & destroy specified window.
    */
   XDestroyWindow (x11_display, win_data->window);

   /***
    * Poziv XDestroyImage() ce obrisati sve sto treba, znaci i zauzetu memoriju
    * na koju pokazuje win_data->shared_memory i strukturu win_data->ximage.
    */
   if (win_data->ximage)
      XDestroyImage (win_data->ximage);

   if (win_data->use_mitshm)
   {
      XShmDetach (x11_display, &win_data->shm_info);
      XSync (x11_display, False);

      if (shmdt (win_data->shm_info.shmaddr) < 0)
         debug (("X shmdt(): %s", strerror (errno)))
   }

   (void) memset (win_data, 0, sizeof (struct WINDOW_DATA));

   debug_endfunc
}


#ifdef ANSI_FUNC

short X11_window_init_mitshm (struct WINDOW_DATA *win_data)
#else
   
short X11_window_init_mitshm (win_data)
   struct WINDOW_DATA *win_data;
#endif
{
   XWindowAttributes win_attr;

   debug (("{%s(0x%08lx)", __FUNCTION__, (long) win_data))

   if (XGetWindowAttributes (x11_display, win_data->window, &win_attr) != True)
      debug_return ("} == %d", FUNC_RET_ERROR)

   win_data->shared_len = win_attr.width * win_attr.height * x11_bpp;

   /***
    * MIT shared memory.
    */
   if (win_data->use_mitshm)
   {
      if ((win_data->use_mitshm = XShmQueryExtension (x11_display)))
      {
         debug (("win_data->shared_len = %ld\n", win_data->shared_len))

         win_data->ximage     = XShmCreateImage (x11_display, x11_visual_info.visual, x11_visual_info.depth, ZPixmap, NULL, &win_data->shm_info, win_attr.width, win_attr.height);
         win_data->shared_len = win_data->ximage->bytes_per_line * win_attr.height;

         debug (("win_data->shared_len = %ld\n", win_data->shared_len))

         /***
          * Get our shared screen memory.
          */
         if ((win_data->shm_info.shmid = shmget (IPC_PRIVATE, win_data->shared_len, IPC_CREAT | 0777)) < 0)
         {
            debug (("Couldn't get X shared memory.\n"))

            win_data->use_mitshm = 0;
         }
         else if ((win_data->shm_info.shmaddr = win_data->ximage->data = (char *) shmat (win_data->shm_info.shmid, NULL, 0)) == (char *) -1)
         {
            debug (("Couldn't attach to X shared memory.\n"))

            (void) shmctl (win_data->shm_info.shmid, IPC_RMID, NULL);

            win_data->use_mitshm = 0;
         }
         else
         {
            win_data->shared_memory     = (unchar *) win_data->shm_info.shmaddr;
            win_data->shm_info.readOnly = False;

            XShmAttach (x11_display, &win_data->shm_info);

            if (shmctl (win_data->shm_info.shmid, IPC_RMID, NULL) < 0)
               debug (("X: shmctl(IPC_RMID): %s\n", strerror (errno)))

            x11_shm_completion = XShmGetEventBase (x11_display) + ShmCompletion;
            x11_shm_busy       = 0;
         }
      }
      else
         debug (("X: MITSHM extension not enabled!\n"))
   }

   if (! win_data->use_mitshm)
   {
      win_data->shared_memory = (unchar *) malloc (win_data->shared_len);
      win_data->ximage        = XCreateImage (x11_display, x11_visual_info.visual, x11_visual_info.depth, ZPixmap, 0, (char *) win_data->shared_memory, win_attr.width, win_attr.height, 8 * x11_bpp, x11_bpp * win_attr.width);
   }

   debug (("use_mitshm = %d\n", win_data->use_mitshm))

   XSync (x11_display, False);

   debug_return ("} == %d", FUNC_RET_OK)
}


#ifdef ANSI_FUNC

void X11_copyscreen (struct WINDOW_DATA *win_data)
#else
   
void X11_copyscreen (win_data)
   struct WINDOW_DATA *win_data;
#endif
{
   XWindowAttributes win_attr;

   debug (("{%s(0x%08lx)", __FUNCTION__, (long) win_data))

   if (XGetWindowAttributes (x11_display, win_data->window, &win_attr) != True)
      debug_noreturn

   if (win_data->use_mitshm)
   {
      while (x11_shm_busy)
         X11_handle_events ();

      XShmPutImage (x11_display, win_data->window, win_data->gc, win_data->ximage, 0, 0, 0, 0, win_attr.width, win_attr.height, True);

      x11_shm_busy = 1;
   }
   else
      XPutImage (x11_display, win_data->window, win_data->gc, win_data->ximage, 0, 0, 0, 0, win_attr.width, win_attr.height);

   XSync (x11_display, False);

   debug_endfunc
}


#ifdef ANSI_FUNC

void X11_draw_line (struct WINDOW_DATA *win_data, short x1, short y1, short x2, short y2, short color_index)
#else

void X11_draw_line (win_data, x1, y1, x2, y2, color_index)
   struct WINDOW_DATA *win_data;
   short x1;
   short y1;
   short x2;
   short y2;
   short color_index;
#endif
{
   debug (("{%s(0x%08lx, %d, %d, %d, %d, %d)", (long) win_data, x1, y1, x2, y2, color_index))

   if (win_data->use_mitshm)
   {
   }
   else
      ;

   debug_endfunc
}


#ifdef ANSI_FUNC

void X11_draw_rectangle (struct WINDOW_DATA *win_data, short x, short y, short width, short height, short color_index)
#else
   
void X11_draw_rectangle (win_data, x, y, width, height, color_index)
   struct WINDOW_DATA *win_data;
   short x;
   short y;
   short width;
   short height;
   short color_index;
#endif
{
   short x2 = x + width - 1, y2 = y + height - 1;

   debug (("{%s(0x%08lx, %d, %d, %d, %d, %d)", (long) win_data, x, y, width, height, color_index))

   X11_draw_line (win_data, x, y, x2, y, color_index);
   X11_draw_line (win_data, x, y2, x2, y2, color_index);
   X11_draw_line (win_data, x, y, x, y2, color_index);
   X11_draw_line (win_data, x2, y, x2, y2, color_index);

   debug_endfunc
}


#ifdef ANSI_FUNC

short X11_save_snapshot (Window win)
#else
   
short X11_save_snapshot (win)
   Window win;
#endif
{
   struct IMAGE_DATA *snapshot;
   XWindowAttributes  win_attr;

   debug (("{%s(0x%08x)", __FUNCTION__, (int) win))

   if (XGetWindowAttributes (x11_display, win, &win_attr) != True)
      debug_return ("} == %d", FUNC_RET_ERROR)

   if ((snapshot = image_alloc (win_attr.width, win_attr.height)) != NULL)
   {
      char filename[MAXPATHLEN];

      (void) sprintf (filename, "snapshot_%ld.pcx", (long) time (NULL));
#if 0  /* XXX  Out of order */
      gl_getbox (0, 0, snapshot->width, snapshot->height, snapshot->image);
      (void) memcpy (snapshot->palette, palette, SIZEOF_PALETTE);
#endif
      (void) pcx_image_save (snapshot, filename);
      image_free (&snapshot);

      debug_return ("} == %d", FUNC_RET_OK)
   }

   debug_return ("} == %d", FUNC_RET_ERROR)
}
