/***
 * Copyright (C) 1996-98. by Miroslav Zagorac <zaga@tirnanog.cc.fer.hr>
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sysexits.h>
#include <sys/param.h>

#ifdef HAVE_GETOPT_LONG
#  include <getopt.h>
#else
#  include "../lib/getopt.h"
#endif

#include "../data/graphics/bd_icon.xpm"
#include "../data/objects_def.h"
#include "../lib/lib_atexit.h"
#include "../lib/lib_debug.h"
#include "../lib/lib_options.h"

#include "cave.h"
#include "cave_data.h"
#include "demo.h"
#include "graphics.h"
#include "keyboard.h"
#include "level_editor.h"
#include "logo.h"
#include "main.h"
#include "objects.h"
#include "player.h"
#include "score.h"
#include "version.h"


struct GAME_DATA    game_data;
struct PLAYER_DATA  player_data[MAX_PLAYERS];
struct FONT_DATA   *bd_font_8x8, *bd_font_16x16;
char               *gamelibdir = GAMELIBDIR;
short               _options = 0;

#ifdef USE_X11
struct WINDOW_DATA main_window;
#endif


#ifdef ANSI_FUNC

static void my_exit (void)
#else

static void my_exit ()
#endif
{
   debug (("{%s()", __FUNCTION__))

#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_SVGALIB)
      svga_close ();
#endif

#ifdef USE_X11
   if (game_data.graphics_type == GRAPHICS_X11)
   {
      X11_window_close (&main_window);
      X11_close ();
   }
#endif

   font_free (&bd_font_8x8);
   font_free (&bd_font_16x16);

   debug_close ();

   debug_endfunc
}


#ifdef ANSI_FUNC

static void signal_handler (int signum)
#else

static void signal_handler (signum)
   int signum;
#endif
{
   debug (("{%s(%s  <%d>)", __FUNCTION__, strsignal (signum), signum))

   switch (signum)
   {
      case SIGINT : exit (EX_SOFTWARE);
   }

   if (signal (signum, signal_handler) == SIG_ERR)
      perror_exit ("signal()", -errno);

   debug_endfunc
}


#ifdef ANSI_FUNC

void game_timer (short init_flag)
#else
   
void game_timer (init_flag)
   short init_flag;
#endif
{
   static struct timeval start_time;
   static short          counter;
#ifdef DEBUG
   static short          loop_count = 0;
#endif

   debug (("{%s(%d)", __FUNCTION__, init_flag))

   if (init_flag)
   {
      (void) gettimeofday (&start_time, NULL);

      counter = 0;
   }
   else
   {
      struct timeval current_time;
      time_t         time_tick;

      (void) gettimeofday (&current_time, NULL);

      time_tick = (current_time.tv_sec - start_time.tv_sec) * 1000000L + current_time.tv_usec - start_time.tv_usec;

      debug (("time_tick = %ld\n", time_tick))

      if (time_tick > 50000L)
      {
         if (counter++ >= 2)
         {
            game_data.game_status |= GAME_OBJECT_MOVE;

            counter = 0;
         }

         if ((game_data.game_status & GAME_PLAY_MODE) && !(game_data.game_status & (GAME_PAUSE | GAME_OUT_OF_TIME | GAME_NEXT_CAVE | GAME_QUIT_CAVE)))
         {
            cave.time        -= time_tick;
            cave.amoeba_time -= time_tick;

            if (game_data.game_status & GAME_MAGIC_WALL)
               cave.wall_time -= time_tick;

            start_time.tv_sec  = current_time.tv_sec;
            start_time.tv_usec = current_time.tv_usec;

#ifdef DEBUG
            if ((loop_count >= 0) && ++loop_count && (cave.time <= 0))
            {
               debug (("loop_count = %d\n", loop_count))

               loop_count = -1;
            }
#endif
         }
         else
            (void) gettimeofday (&start_time, NULL);

         game_data.game_status |= GAME_OBJECT_ANIMATE;
      }
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

time_t usec_to_sec (long usec)
#else
   
time_t usec_to_sec (usec)
   long usec;
#endif
{
   return ((usec > 0) ? (usec / 1000000L + ((usec % 1000000L) ? 1 : 0)) : 0);
}


#ifdef ANSI_FUNC

static void top_print (void)
#else

static void top_print ()
#endif
{
   debug (("{%s()", __FUNCTION__))

   if (! (game_data.game_status & GAME_PLAY_MODE))
   {
      if (code_room (player_data[game_data.current_player].room) < 'A')
         graph_printf (4 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGRAY, bd_font_16x16, "PLAYER %d,   %d %s,   BONUS LEVEL", game_data.current_player + 1, player_data[game_data.current_player].lives, (player_data[game_data.current_player].lives > 1) ? "MEN" : "MAN");
      else
         graph_printf (3 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGRAY, bd_font_16x16, "  PLAYER %d,   %d %s,   ROOM %c/%02d  ", game_data.current_player + 1, player_data[game_data.current_player].lives, (player_data[game_data.current_player].lives > 1) ? "MEN" : "MAN", code_room (player_data[game_data.current_player].room), player_data[game_data.current_player].level + 1);
   }
   else if (cave.diamonds)
   {
      graph_printf (4 * CHAR_16x16, 0.5 * CHAR_16x16, RED, bd_font_16x16, "%02d", cave.diamonds);
      graph_printf (6 * CHAR_16x16, 0.5 * CHAR_16x16, WHITE, bd_font_16x16, "\023");
      graph_printf (7 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGRAY, bd_font_16x16, "%02d  ", cave.dvalue_init);
      graph_printf (11 * CHAR_16x16, 0.5 * CHAR_16x16, GREEN, bd_font_16x16, "%02d  ", game_data.diamonds);
      graph_printf (15 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGRAY, bd_font_16x16, "%03ld  %06ld  PLAYER:%d", usec_to_sec (cave.time), player_data[game_data.current_player].score, game_data.current_player + 1);
   }
   else
   {
      graph_printf (4 * CHAR_16x16, 0.5 * CHAR_16x16, WHITE, bd_font_16x16, "\023\023\023");
      graph_printf (7 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGRAY, bd_font_16x16, "%02d  ", cave.dvalue_init);
      graph_printf (11 * CHAR_16x16, 0.5 * CHAR_16x16, GREEN, bd_font_16x16, "%02d  ", game_data.diamonds);
      graph_printf (15 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGRAY, bd_font_16x16, "%03ld  %06ld  PLAYER:%d", usec_to_sec (cave.time), player_data[game_data.current_player].score, game_data.current_player + 1);
   }

   debug_endfunc
}


#ifdef DEBUG

#ifdef ANSI_FUNC

void show_refresh_rate (short init_flag)
#else

void show_refresh_rate (init_flag)
   short init_flag;
#endif
{
   static struct timeval start_time, current_time;
   static ulong          frames;

   if (init_flag)
   {
      frames = 0;

      (void) gettimeofday (&start_time, NULL);
   }
   else
   {
      (void) gettimeofday (&current_time, NULL);

      graph_printf (36 * CHAR_16x16, 0.5 * CHAR_16x16, BLUE, bd_font_16x16, "%4.0f", 10.0 * ++frames / ((current_time.tv_sec - start_time.tv_sec) + (current_time.tv_usec - start_time.tv_usec) / 1000000.0));
   }
}

#endif /* DEBUG */


#ifdef ANSI_FUNC

short game (void)
#else

short game ()
#endif
{
   short pause_counter = 0;

   debug (("{%s()", __FUNCTION__))

   if (cave_load (player_data[game_data.current_player].room, player_data[game_data.current_player].level) != FUNC_RET_OK)
      debug_return ("} == %d", FUNC_RET_ERROR)

   top_print ();
   cave_mask_set (CAVE_MASK_DEC | CAVE_MASK_FADE, INIT_TIME);
   show_refresh_rate (1);
   game_timer (1);

   cave.data[CAVE_POSITION (rockford_coord.x, rockford_coord.y)] = PRE_ROCKFORD_2;

   while (1)
   {
      game_timer (0);
      (void) key_update ();

      if (game_data.game_status & GAME_OBJECT_ANIMATE)
      {
         if (game_data.game_status & GAME_OBJECT_MOVE)
         {
            if (game_data.game_status & GAME_PAUSE)
            {
               game_data.scroll_step = game_data.scroll_basic_step * 2;

               if (pause_counter++ & 0x10)
                  top_print ();
               else
                  graph_printf (3.5 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGREEN, bd_font_16x16, "    PRESS SPACE KEY TO RESUME    ");

               if (key_pressed_code (KEY_SPACE, 0))
               {
                  game_data.game_status &= ~GAME_PAUSE;
                  game_data.scroll_step  = game_data.scroll_basic_step;
               }
            }
            else
            {
               pause_counter = 0;

               if (usec_to_sec (cave.time) <= 0L)
               {
                  game_data.game_status |= GAME_OUT_OF_TIME;

                  debug_return ("} == %d", game_data.game_status)
               }

               top_print ();

               if (game_data.game_status & GAME_PLAY_MODE)
               {
                  objects_explode ();
                  objects_fall ();

                  /***
                   * Ako je Rockford mrtav, onda se iz igre izlazi sa pritiskom na FIRE tipku.
                   * U slucaju da je jos uvijek ziv, testira se da li je dosao do izlaza,
                   * te ako ni to nije, da li su pritisnute neke druge tipke za gubljenje
                   * zivota i ostalog.
                   */
                  if (game_data.game_status & GAME_PLAYER_DEAD)
                  {
                     if (key_pressed_code (KEY_GAME_FIRE, 0))
                     {
                        game_data.game_status |= GAME_QUIT_CAVE;

                        debug_return ("} == %d", game_data.game_status)
                     }
                  }
                  else if (player_move ())
                  {
                     game_data.game_status |= GAME_NEXT_CAVE;

                     debug_return ("} == %d", game_data.game_status)
                  }

                  if (key_pressed_code (KEY_GAME_QUIT, 0))
                  {
                     game_data.game_status |= GAME_QUIT_CAVE;

                     debug_return ("} == %d", game_data.game_status)
                  }
                  else if (key_pressed_code (KEY_GAME_RESET, 0))
                  {
                     game_data.game_status |= GAME_RESET;

                     debug_return ("} == %d", game_data.game_status)
                  }
                  else if ((game_data.game_status & GAME_DEMO_MODE) && key_pressed_code (KEY_GAME_FIRE, 0))
                  {
                     game_data.game_status |= GAME_QUIT_CAVE;

                     debug_return ("} == %d", game_data.game_status)
                  }

                  /***
                   * Da li je sakupljen dovoljan broj dijamanata?
                   */
                  if (game_data.diamonds == cave.diamonds)
                  {
                     cave.diamonds    = 0;
                     cave.dvalue_init = cave.dvalue_extra;

                     game_data.game_status |= GAME_DIAMONDS_COLLECTED;

                     cave.data[CAVE_POSITION (cave_out.x, cave_out.y)] = FLASH_OUTBOX;

                     graph_setpalettecolor (0, 255, 255, 255);
                  }

                  objects_move ();
               }
            }

            game_data.game_status &= ~GAME_OBJECT_MOVE;
         }

         objects_animate ();
      }

      cave_show ();
      graph_copyscreen ();

      if (game_data.game_status & GAME_DIAMONDS_COLLECTED)
      {
         graph_setpalettecolor (0, 0, 0, 0);

         game_data.game_status &= ~GAME_DIAMONDS_COLLECTED;
      }

      show_refresh_rate (0);
   }

   debug_return ("} == %d", game_data.game_status)
}


#ifdef ANSI_FUNC

static void usage (char *program_name)
#else

static void usage (program_name)
   char *program_name;
#endif
{
   debug (("{%s(\"%s\")", __FUNCTION__, program_name))

   (void) printf ("\nusage: %s {--help -h}\n", program_name);
   (void) printf ("       %s {--version -V}\n", program_name);
   (void) printf ("       %s [OPTION]...\n\n", program_name);

   if (TEST_OPTION (OPT_HELP))
   {
      (void) printf ("Options are:\n");

      (void) printf ("  -c, --commodore-graphics   Use commodore 64 graphics.\n");
#ifdef DEBUG
      (void) printf ("  -C, --check-cave-data      Check all caves data.\n");
      (void) printf ("  -d, --debug                Enable debugging options.\n");
#endif
      (void) printf ("  -h, --help                 Show this text and exit.\n");
      (void) printf ("  -l, --game-libdir          Change game library directory\n\t\t\t\t(default: %s).\n", GAMELIBDIR);
#ifdef USE_X11
      (void) printf ("  -n, --no-mitshm            Don't use the MIT shared memory extension.\n");
#endif
      (void) printf ("  -s, --spectrum-graphics    Use ZX spectrum graphics.\n");
      (void) printf ("  -V, --version              Show program version and exit.\n");
      (void) printf ("  -v, --verbose              Increase verbose level.\n");
#ifdef USE_SVGALIB
      (void) printf ("  -w, --dont-wait-retrace    Don't wait for wertical retrace (SVGAlib only).\n\n");
#endif
      (void) printf ("\n");
   }
   else
      (void) printf ("For help type: %s -h\n\n", program_name);

   Exit (EX_USAGE);
}


#ifdef ANSI_FUNC

void game_over (char *title)
#else
   
void game_over (title)
   char *title;
#endif
{
   unchar pause_counter = 0;

   debug (("{%s(\"%s\")", __FUNCTION__, title))

   do
   {
      game_timer (0);

      if (game_data.game_status & GAME_OBJECT_ANIMATE)
      {
         if ((pause_counter++ & 0x20) || !title)
            top_print ();
         else
            graph_printf (3.5 * CHAR_16x16, 0.5 * CHAR_16x16, LIGHTGRAY, bd_font_16x16, title);

         objects_animate ();
      }

      cave_show ();
      graph_copyscreen ();

      (void) key_update ();
   }
   while (!key_pressed_code (KEY_GAME_FIRE, 0) && !key_pressed_code (KEY_SPACE, 0));

   debug_endfunc
}


#ifdef ANSI_FUNC

void time_bonus (void)
#else
   
void time_bonus ()
#endif
{
   debug (("{%s()", __FUNCTION__))

   while (usec_to_sec (cave.time))
   {
      cave.time -= 1000000L;

      top_print ();

      if (game_data.game_status & GAME_OBJECT_ANIMATE)
         objects_animate ();

      cave_show ();
      score_add (1);
      graph_copyscreen ();
   }

   debug_endfunc
}


#ifdef ANSI_FUNC

int main (int argc, char **argv, char **envp)
#else

int main (argc, argv, envp)
   int argc;
   char **argv;
   char **envp;
#endif
{
   static struct option longopts[] =
   {
      { "commodore-graphics", 0, NULL, 'c'  },
#ifdef DEBUG
      { "check-cave-data"   , 0, NULL, 'C'  },
      { "debug"             , 0, NULL, 'd'  },
#endif
      { "help"              , 0, NULL, 'h'  },
      { "game-libdir"       , 1, NULL, 'l'  },
#ifdef USE_X11
      { "no-mitshm"         , 0, NULL, 'n'  },
#endif
      { "spectrum-graphics" , 0, NULL, 's'  },
      { "version"           , 0, NULL, 'V'  },
      { "verbose"           , 0, NULL, 'v'  },
#ifdef USE_SVGALIB
      { "dont-wait-retrace" , 0, NULL, 'w'  },
#endif
      { NULL                , 0, NULL, '\0' }
   };
   int   longind = 0;
   short error = 0, option;

   debug_main_arg
   debug (("GETOPT_ARGS = \"%s\"\n", GETOPT_ARGS))

   use_perror = 1;

   /*
    * Ovdje testiram samo flagove '--help -h' i '--version -V'.
    * Ako je neki od njih postavljen, program ispise odgovarajucu poruku i prestaje sa radom.
    */
   for (opterr = 0; (option = getopt_long (argc, argv, GETOPT_ARGS, longopts, &longind)) != EOF; )
      switch (option)
      {
         case 'h' : _options |= OPT_HELP;
         
                    usage (*argv);

                    break;

         case 'V' : (void) printf ("Boulder Dash V%s by %s, %s\n", _VERSION, _AUTHOR, __DATE__);

                    debug_return ("} == %d", EX_OK)
      }

   for (opterr = 1, optind = 0; (option = getopt_long (argc, argv, GETOPT_ARGS, longopts, &longind)) != EOF; )
      switch (option)
      {
         case 'c' : if (TEST_OPTION (OPT_SPECTRUM))
                    {
                    }
                    else
                       _options |= OPT_COMMODORE;

                    break;

#ifdef DEBUG
         case 'C' : _options |= OPT_CHECK_CAVEDATA;

                    break;

         case 'd' : if (! TEST_OPTION (OPT_DEBUG))
                    {
                       _options |= OPT_DEBUG;

                       debug_open ("/tmp/BoulderDash.log");
                    }

                    break;
#endif

         case 'h' : break;

         case 'l' : if (TEST_OPTION (OPT_GAMELIBDIR))
                    {
                       (void) printf ("The Boulder Dash libdir is already defined!\n");

                       error++;
                    }
                    else
                    {
                       _options |= OPT_GAMELIBDIR;

		       gamelibdir = optarg;
                    }

                    break;

#ifdef USE_X11
         case 'n' : _options |= OPT_NOMITSHM;

                    break;
#endif

         case 's' : if (TEST_OPTION (OPT_COMMODORE))
                    {
                    }
                    else
                       _options |= OPT_SPECTRUM;

                    break;

         case 'V' : break;

         case 'v' : _options |= OPT_VERBOSE;
                    verbose_level++;
                    
                    break;

#ifdef USE_SVGALIB
         case 'w' : _options |= OPT_WAITRETRACE;

                    break;
#endif

         case '?' : error++;
      }

   if (optind < argc)
   {
      error++;

      (void) printf ("Too many arguments:");

      for ( ; optind < argc; optind++)
         (void) printf (" \"%s\"", argv[optind]);

      (void) printf (".\n");
   }

   if (error)
      usage (*argv);

   if (! TEST_OPTION (OPT_DEBUG))
      debug_off ();

#ifdef DEBUG
   if (TEST_OPTION (OPT_CHECK_CAVEDATA))
      caves_all_check ();
#endif

   if (! (bd_font_8x8 = font_load (gamelibdir, "font_8x8.pcx")))
      perror_exit ("font_8x8.pcx", -errno);

   if (! (bd_font_16x16 = font_load (gamelibdir, "font_16x16.pcx")))
      perror_exit ("font_16x16.pcx", -errno);

   bd_font_16x16->writemode = WRITEMODE_OVERWRITE | FONT_EXPANDED;

   game_data.scroll_basic_step = SCROLL_BASIC_STEP;
   game_data.graphics_type     = GRAPHICS_UNKNOWN;

#ifdef USE_X11
   main_window.attr.backing_store           = Always;
   main_window.attr_mask                    = CWBackingStore;
   main_window.gc_values.graphics_exposures = False;
   main_window.gc_values_mask               = GCGraphicsExposures;
   main_window.window_name                  = "Boulder Dash";
   main_window.icon_name                    = "Boulder Dash";
   main_window.icon                         = bd_icon_xpm;
   main_window.event_mask                   = KeyPressMask | KeyReleaseMask | StructureNotifyMask | EnterWindowMask | LeaveWindowMask;
   main_window.use_mitshm                   = TEST_OPTION (OPT_NOMITSHM) ? 0 : 1;

   if (X11_init () == FUNC_RET_OK)
   {
      (void) setuid (getuid ());
      (void) setgid (getgid ());

      X11_callback_keyevent = key_callback;

      if (X11_set_colormap (bd_palette, X11_MAX_COLORS) != FUNC_RET_OK)
         perror_exit ("Cannot allocate colormap!", EX_SOFTWARE);

      if (X11_window_open (&main_window, WINDOW_WIDTH, WINDOW_HEIGHT) != FUNC_RET_OK)
         perror_exit ("Cannot open window!", EX_SOFTWARE);

      if (X11_window_init_mitshm (&main_window) != FUNC_RET_OK)
         perror_exit ("Cannot allocate memory!", EX_SOFTWARE);

      debug (("main_window = {\n"))
      debug (("   ximage        = 0x%08lx\n", (long) main_window.ximage))
      debug (("   shm_info      = 0x%08x\n", main_window.shm_info))
      debug (("   use_mitshm    = %d\n", main_window.use_mitshm))
      debug (("   shared_len    = %ld\n", main_window.shared_len))
      debug (("   shared_memory = 0x%08lx\n", (long) main_window.shared_memory))
      debug ((" }\n"))

      /***
       * Malo cu pricekati da se prozor mapira...
       */
      do
         XNextEvent (x11_display, &main_window.event);
      while (main_window.event.type != MapNotify);

      XFlush (x11_display);

      debug (("X Windows system initializated!\n"))

      if (x11_bpp > 1)
         game_data.scroll_basic_step *= 2;

      game_data.graphics_type = GRAPHICS_X11;
   }
#  ifndef USE_SVGALIB
   else
      perror_exit ("Cannot open display!\n", EX_SOFTWARE);
#  endif
#endif /* USE_X11 */

#ifdef USE_SVGALIB
   if (game_data.graphics_type == GRAPHICS_UNKNOWN)
   {
      if (svga_init (VGAMODE, 0) != FUNC_RET_OK)
         perror_exit ("Cannot init SVGALIB", -errno);

      (void) setuid (getuid ());
      (void) setgid (getgid ());

      if (svga_virtual_init (VGAMODE, 1) != FUNC_RET_OK)
         perror_exit ("Cannot initialize virtual buffers", EX_SOFTWARE);

      svga_set_palette (bd_palette, 1);

      game_data.graphics_type = GRAPHICS_SVGALIB;
   }
#endif /* USE_SVGALIB */

   (void) atexit (my_exit);

   if (signal (SIGINT, signal_handler) == SIG_ERR)
      perror_exit ("signal(SIGINT)", -errno);
                             
   while (1)
   {
      ushort game_retval = FUNC_RET_OK;

      (void) logo_show ();
      coords_clear (&cave_offset);
      player_data_set (MAX_PLAYERS, game_data.start_room, game_data.start_level);

      game_data.current_player = 0;

      while (player_data[game_data.current_player].lives)
      {
         if (((game_retval = game ()) & ~GAME_STATUS_MASK) != FUNC_RET_OK)
            break;
         else if (game_retval & GAME_NEXT_CAVE)
         {
            time_bonus ();
            cave_next ();
         }
         else if (game_retval & GAME_QUIT_CAVE)
         {
            if (code_room (player_data[game_data.current_player].room) < 'A')
               cave_next ();
            else if (! --player_data[game_data.current_player].lives)
               game_over ("        G A M E   O V E R        ");
         }
         else if (game_retval & GAME_OUT_OF_TIME)
         {
            if (code_room (player_data[game_data.current_player].room) < 'A')
               cave_next ();
            else
               player_data[game_data.current_player].lives--;

            game_over (player_data[game_data.current_player].lives ? "      O U T   O F   T I M E      " : "        G A M E   O V E R        ");
         }
         else if (game_retval & GAME_RESET)
            break;

         cave_mask_set (CAVE_MASK_INC, INIT_TIME);
      }

      if ((game_retval & ~GAME_STATUS_MASK) == FUNC_RET_OK)
         graph_palette_fade (-4);
   }

   debug_return ("} == %d", EX_OK)
}
